package de.hft.stuttgart.citydoctor2.zip;

import de.hft.stuttgart.citydoctor2.datastructure.CityDoctorModel;
import de.hft.stuttgart.citydoctor2.parser.CityGmlParseException;
import de.hft.stuttgart.citydoctor2.parser.CityGmlParser;
import de.hft.stuttgart.citydoctor2.parser.InvalidGmlFileException;
import de.hft.stuttgart.citydoctor2.parser.ParserConfiguration;
import de.hft.stuttgart.citydoctor2.parser.ProgressListener;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import java.io.IOException;
import java.io.Serial;
import java.io.Serializable;
import java.nio.file.Path;
import java.util.zip.ZipEntry;

/**
 * This class serves as a representation of a CityGml file in a ZIP archive.
 * <p/>
 * The {@link CityDoctorModel} of the associated file can only be accessed after loading the corresponding entry.
 */
public class CityGmlZipEntry implements Serializable {

    private static final Logger logger = LogManager.getLogger(CityGmlZipEntry.class);
    @Serial
    private static final long serialVersionUID = -5732913269959043262L;

    private final Path filePath;
    private final String fileName;
    private final String displayName;
    private transient CityDoctorModel model;
    private CityGmlZipArchive parentArchive;
    private boolean inMemory;
    private long fileSize = -1L;
    private static final long MB = 1024 * 1024L;
    private ZipEntryErrorType errorType = null;


    /**
     * Attempts to parse the CityGML model associated with this entry and loads it into memory.
     * <p/>
     * In the event of a parsing error, the entry will be flagged by assigning the appropriate
     * {@link ZipEntryErrorType} to the errorType field.
     *
     * @param config Parser configuration for this model
     */
    public void loadEntry(ParserConfiguration config) {
        loadEntry(config, null);
    }

    /**
     * Attempts to parse the CityGML model associated with this entry and loads it into memory.
     * <p/>
     * In the event of a parsing error, the entry will be flagged by assigning the appropriate
     * {@link ZipEntryErrorType} to the errorType field.
     *
     * @param config Parser configuration for this model
     * @param l      ProgressListener for tracking parsing progress
     */
    public void loadEntry(ParserConfiguration config, ProgressListener l) {
        if (inMemory) {
            return;
        }
        if (errorType != null) {
            logger.warn("Tried loading erroneous CityGmlZipEntry");
            return;
        }
        try {
            this.model = CityGmlParser.parseCityGmlZipEntry(this, config, l);
            this.inMemory = true;
        } catch (CityGmlParseException | InvalidGmlFileException e) {
            logger.error(e);
            this.errorType = ZipEntryErrorType.INVALID_CITY_GML_FILE;
        } catch (IOException e) {
            logger.error(e);
            this.errorType = ZipEntryErrorType.IO_ERROR;
        }
    }

    /**
     * Registers a {@link ZipEntry ZipEntry} in a ZIP-archive as a {@link CityGmlZipEntry}.
     * <p/>
     * If the filesize of the associated file exceeds the total available system memory, or read access is denied,
     * the entry will be flagged by assigning the appropriate {@link ZipEntryErrorType} to the errorType field.
     *
     * @param entry         The ZipEntry object
     * @param parentArchive The parent archive for this entry
     * @return A corresponding CityGmlZipEntry object
     */
    public static CityGmlZipEntry register(ZipEntry entry, CityGmlZipArchive parentArchive) {
        CityGmlZipEntry cgzEntry = new CityGmlZipEntry(entry, parentArchive, false);
        try {
            if (!cgzEntry.entrySizeWithinMemoryLimits()) {
                cgzEntry.errorType = ZipEntryErrorType.EXCESSIVE_FILESIZE;
            }
        } catch (IOException e) {
            logger.error(e);
            cgzEntry.errorType = ZipEntryErrorType.IO_ERROR;
        }
        return cgzEntry;
    }

    /**
     * Checks if the filesize of this entry would exceed the system's available memory.
     *
     * @return true if within limits, false otherwise
     * @throws IOException if read access to the ZIP-archive is denied
     */
    private boolean entrySizeWithinMemoryLimits() throws IOException {
        long memoryLimit = (long) Math.ceil(((double) Runtime.getRuntime().maxMemory() / MB) * 0.9);
        if (fileSize == -1L) {
            try (CityGmlZipEntryFile entryFile = new CityGmlZipEntryFile(this)) {
                long filesize = entryFile.getFileSize();
                if (filesize != -1) {
                    this.fileSize = filesize;
                } else {
                    return false;
                }
            } catch (Exception e) {
                throw new IOException(e);
            }
        }
        return memoryLimit > (fileSize / MB);
    }

    private CityGmlZipEntry(ZipEntry entry, CityGmlZipArchive parentArchive, boolean inMemory) {
        this.fileName = entry.getName();
        this.filePath = Path.of(entry.getName());
        this.displayName = filePath.getFileName().toString();
        if (entry.getSize() != -1) {
            this.fileSize = entry.getSize();
        }
        this.model = null;
        this.inMemory = inMemory;
        this.parentArchive = parentArchive;
    }

    /**
     * Gets the parent archive of this entry.
     * @return the archive
     */
    public CityGmlZipArchive getArchive() {
        return parentArchive;
    }

    /**
     * Gets the sub-path in the archive of the file associated with this entry.
     *
     * @return the relative sub-path from the archive root directory
     */
    public String getEntrySubPath() {
        return fileName;
    }

    /**
     * Gets the name of the file associated with this entry.
     * @return the filename
     */
    public String getDisplayName() {
        return displayName;
    }

    /**
     * Gets the {@link ZipEntryErrorType} of this entry.
     * @return the error type, or null if this entry is not erroneous
     */
    public ZipEntryErrorType getErrorType() {
        return errorType;
    }

    /**
     * Gets the {@link CityDoctorModel} of this entry.
     * @return The model, or null if the entry has not been loaded or is erroneous
     */
    public CityDoctorModel getModel() {
        return model;
    }

    public void setFileSize(long size) {
        fileSize = size;
    }

    /**
     * Resolves the reference path of a {@link de.hft.stuttgart.citydoctor2.datastructure.LibraryObject LibraryObject}
     * in the {@link CityDoctorModel} of this entry.
     * @param libraryObjectPath The reference path to the LibraryObject's CityGml file
     * @return Relative sub-path within the archive to the CityGml file
     */
    public Path resolveLibraryObjectPath(String libraryObjectPath) {
        return filePath.getParent().resolve(libraryObjectPath);
    }

    /**
     * Checks if the {@link CityDoctorModel} of this entry has been loaded into memory.
     *
     * @return true if it is loaded, false otherwise
     */
    public boolean isLoaded() {
        return inMemory;
    }

    /**
     * Returns the estimated uncompressed filesize of this entry.
     *
     * @return filesize in bytes, -1L if filesize is unknown
     */
    public long getFileSize() {
        return fileSize;
    }
}
